import malevich_coretools.funcs.funcs as f
from malevich_coretools.abstract import *  # noqa: F403
from malevich_coretools.batch import Batcher
from malevich_coretools.secondary import Config


def admin_get_runs(
    *,
    auth: Optional[AUTH] = None,
    conn_url: Optional[str] = None,
    batcher: Optional[Batcher] = None,
) -> AdminRunsInfo:
    """return info about all runs"""
    if batcher is None:
        batcher = Config.BATCHER
    if batcher is not None:
        return batcher.add("getAllRuns", result_model=AdminRunsInfo)
    return f.get_admin_runs(auth=auth, conn_url=conn_url)


def admin_get_run_info(
    id: Alias.Id,
    *,
    auth: Optional[AUTH] = None,
    conn_url: Optional[str] = None,
    batcher: Optional[Batcher] = None,
) -> Alias.Json:
    """return run info by operation `id`"""
    if batcher is None:
        batcher = Config.BATCHER
    data = OperationOrNone(operationId=id)
    if batcher is not None:
        return batcher.add("getRunInfo", data=data)
    return f.get_admin_runs_info(data, auth=auth, conn_url=conn_url)


def admin_get_runs_info(
    *,
    auth: Optional[AUTH] = None,
    conn_url: Optional[str] = None,
    batcher: Optional[Batcher] = None,
) -> Alias.Json:
    """return json: dm id -> info about its runs"""
    if batcher is None:
        batcher = Config.BATCHER
    data = OperationOrNone()
    if batcher is not None:
        return batcher.add("getRunInfo", data=data)
    return f.get_admin_runs_info(data, auth=auth, conn_url=conn_url)


def admin_update_superuser(
    login: str,
    is_superuser: bool = True,
    *,
    auth: Optional[AUTH] = None,
    conn_url: Optional[str] = None,
    batcher: Optional[Batcher] = None,
) -> Alias.Info:
    """update user role by login"""
    if batcher is None:
        batcher = Config.BATCHER
    data = Superuser(login=login, isSuperuser=is_superuser)
    if batcher is not None:
        return batcher.add("postUpdateSuperuser", data=data)
    return f.post_admin_update_superuser(data, auth=auth, conn_url=conn_url)


def admin_delete_run(
    id: Alias.Id,
    withLogs: bool = False,
    *,
    auth: Optional[AUTH] = None,
    conn_url: Optional[str] = None,
    batcher: Optional[Batcher] = None,
) -> Alias.Json:
    """delete run by operation `id`"""
    if batcher is None:
        batcher = Config.BATCHER
    data = AdminStopOperation(operationId=id, withLogs=withLogs)
    if batcher is not None:
        return batcher.add("deleteRuns", data=data)
    return f.delete_admin_runs(data, auth=auth, conn_url=conn_url)


def admin_delete_runs(
    withLogs: bool = False,
    *,
    auth: Optional[AUTH] = None,
    conn_url: Optional[str] = None,
    batcher: Optional[Batcher] = None,
) -> Alias.Json:
    """delete all runs"""
    if batcher is None:
        batcher = Config.BATCHER
    data = AdminStopOperation(withLogs=withLogs)
    if batcher is not None:
        return batcher.add("deleteRuns", data=data)
    return f.delete_admin_runs(data, auth=auth, conn_url=conn_url)



import json
from typing import Any, Dict, List, Optional, Union
from uuid import uuid4

import pandas as pd

from malevich_coretools.abstract.abstract import (
    DEFAULT_MSG_URL,
    Alias,
    AppSettings,
    BasePlatformSettings,
    Cfg,
    DocsDataCollection,
    EndpointOverride,
    Restrictions,
    RunSettings,
    ScaleInfo,
    TaskComponent,
    TaskPolicy,
    UserConfig,
)
from malevich_coretools.batch import Batcher
from malevich_coretools.funcs.checks import check_profile_mode
from malevich_coretools.funcs.funcs import (
    post_collections_data,
    post_collections_data_id,
)
from malevich_coretools.secondary import Config, to_json

__all__ = ["create_collection_from_file_df", "update_collection_from_file_df", "raw_collection_from_df", "raw_collection_from_file", "create_collection_from_df", "update_collection_from_df", "create_app_settings", "create_user_config", "create_task_component", "create_task_policy", "create_restrictions", "create_run_settings", "create_endpoint_override", "create_cfg_struct"]


def create_collection_from_file_df(
    file: str,
    name: Optional[str],
    metadata: Optional[Union[Dict[str, Any], str]],
    *args,
    **kwargs
) -> Alias.Id:
    try:
        data = pd.read_csv(file)
    except pd.errors.EmptyDataError:
        data = pd.DataFrame()
    if name is None:
        name = file
    return create_collection_from_df(data, name=file, metadata=metadata, *args, **kwargs)


def update_collection_from_file_df(
    id: str,
    file: str,
    name: Optional[str],
    metadata: Optional[Union[Dict[str, Any], str]],
    *args,
    **kwargs
) -> Alias.Id:
    try:
        data = pd.read_csv(file)
    except pd.errors.EmptyDataError:
        data = pd.DataFrame()
    if name is None:
        name = file
    return update_collection_from_df(id, data, name=file, metadata=metadata, *args, **kwargs)


def raw_collection_from_df(
    data: pd.DataFrame,
    name: Optional[str],
    metadata: Optional[Union[Dict[str, Any], str]],
) -> DocsDataCollection:
    if metadata is not None:
        if isinstance(metadata, str):
            with open(metadata) as f:
                metadata = json.load(f)
            metadata = json.dumps(metadata)
        elif isinstance(data, dict):
            metadata = json.dumps(metadata)
        elif Config.WITH_WARNINGS:
            Config.logger.warning("wrong metadata type, ignore")
            metadata = None
    return DocsDataCollection(data=[row.to_json() for _, row in data.iterrows()], name=name, metadata=metadata)


def raw_collection_from_file(
    file: str,
    name: Optional[str] = None,
    metadata: Optional[Union[Dict[str, Any], str]] = None,
) -> DocsDataCollection:
    try:
        data = pd.read_csv(file)
    except pd.errors.EmptyDataError:
        data = pd.DataFrame()
    return raw_collection_from_df(data, name, metadata)


def create_collection_from_df(
    data: pd.DataFrame,
    name: Optional[str],
    metadata: Optional[Union[Dict[str, Any], str]],
    batcher: Optional[Batcher] = None,
    *args,
    **kwargs
) -> Alias.Id:
    if batcher is None:
        batcher = Config.BATCHER
    data = raw_collection_from_df(data, name, metadata)
    if batcher is not None:
        return batcher.add("postCollectionByDocs", data=data)
    return post_collections_data(data, *args, **kwargs)


def update_collection_from_df(
    id: str,
    data: pd.DataFrame,
    name: Optional[str],
    metadata: Optional[Union[Dict[str, Any], str]],
    batcher: Optional[Batcher] = None,
    *args,
    **kwargs
) -> Alias.Id:
    if batcher is None:
        batcher = Config.BATCHER
    data = raw_collection_from_df(data, name, metadata)
    if batcher is not None:
        return batcher.add("postCollectionByDocsAndId", data=data, vars={"id": id})
    return post_collections_data_id(id, data, *args, **kwargs)


def create_app_settings(
    app_id: str,
    task_id: Optional[str] = None,
    save_collections_name: Union[str, List[str]] = None,
) -> AppSettings:
    if save_collections_name is None:
        gen_collections_name = f"collection-{uuid4()}"
        Config.logger.warning(f"auto generating save collections name for {task_id}${app_id}: {gen_collections_name}")
        save_collections_name = [gen_collections_name]
    elif not isinstance(save_collections_name, list):
        save_collections_name = [save_collections_name]
    return AppSettings(
        appId=app_id, taskId=task_id, saveCollectionsName=save_collections_name
    )


def create_user_config(
    collections: Optional[Dict[str, str]] = None,
    raw_collections: Optional[Dict[str, List[Alias.Doc]]] = None,
    raw_map_collections: Optional[Dict[str, List[Dict[str, Any]]]] = None,
    different: Optional[Dict[str, str]] = None,
    schemes_aliases: Optional[Dict[str, str]] = None,
    msgUrl: Optional[str] = None,
    init_apps_update: Optional[Dict[str, bool]] = None,
    app_settings: Optional[List[AppSettings]] = None,
    app_cfg_extension: Optional[Dict[str, Union[Dict[str, Any], Alias.Json]]] = None,
    email: Optional[str] = None,
) -> UserConfig:
    app_cfg_extension_fixed = {}
    if app_cfg_extension is not None:
        for task_id_app_id, cfg in app_cfg_extension.items():
            app_cfg_extension_fixed[task_id_app_id] = to_json(cfg)
    return UserConfig(
        collections=collections,
        rawCollections=raw_collections,
        rawMapCollections=raw_map_collections,
        different=different,
        schemesAliases=schemes_aliases,
        msgUrl=msgUrl,
        initAppsUpdate=init_apps_update,
        appSettings=app_settings,
        appCfgExtension=app_cfg_extension_fixed,
        email=email,
    )


def create_task_component(
    app_control: Optional[str] = None,
    control: Optional[str] = None,
    extra: Optional[str] = None,
    internal: Optional[str] = None,
    key_value: Optional[str] = None,
    minimal: Optional[str] = None,
    object_storage: Optional[str] = None,
) -> TaskComponent:
    return TaskComponent(
        appControl=app_control,
        control=control,
        extra=extra,
        internal=internal,
        keyValue=key_value,
        minimal=minimal,
        objectStorage=object_storage
    )


def create_task_policy(
    lazy_app_init: bool = False,
    remove_app_after_run: bool = False,
    continue_after_processor: bool = False,
) -> TaskPolicy:
    return TaskPolicy(
        lazyAppInit=lazy_app_init,
        removeAppAfterRun=remove_app_after_run,
        continueAfterProcessor=continue_after_processor,
    )


def create_restrictions(
    honest_scale: bool = True,
    single_pod: bool = False,
    smart_apps_reuse: bool = False,
) -> Restrictions:
    return Restrictions(
        honestScale=honest_scale,
        singlePod=single_pod,
        smartAppsReuse=smart_apps_reuse,
    )


def create_run_settings(
    callback_url: Optional[str] = None,
    debug_mode: bool = False,
    core_manage: bool = False,
    single_request: bool = True,
    wait_runs: bool = True,
    profile_mode: Optional[str] = None,
    scale_info: Optional[List[ScaleInfo]] = None,
    component: Optional[TaskComponent] = None,
    policy: Optional[TaskPolicy] = None,
    restrictions: Optional[Restrictions] = None,
) -> RunSettings:
    check_profile_mode(profile_mode)
    scale_info = [] if scale_info is None else scale_info
    component = create_task_component() if component is None else component
    policy = create_task_policy() if policy is None else policy
    restrictions = create_restrictions() if restrictions is None else restrictions
    return RunSettings(
        callbackUrl=callback_url,
        debugMode=debug_mode,
        coreManage=core_manage,
        singleRequest=single_request,
        waitRuns=wait_runs,
        profileMode=profile_mode,
        scaleInfo=scale_info,
        component=component,
        policy=policy,
        restrictions=restrictions,
    )


def create_endpoint_override(
    cfg_id: Optional[str] = None,
    cfg: Optional[UserConfig] = None,
    override_config: bool = False,
    format_logs: bool = True,
    with_result: bool = True,
    # runSettings part
    callback_url: Optional[str] = None,
    debug_mode: bool = False,
    core_manage: bool = False,
    single_request: bool = True,
    wait_runs: bool = True,
    profile_mode: Optional[str] = None,
    scale_info: Optional[List[ScaleInfo]] = None,
    component: Optional[TaskComponent] = None,
    policy: Optional[TaskPolicy] = None,
    restrictions: Optional[Restrictions] = None,
) -> EndpointOverride:
    return EndpointOverride(
        cfgId=cfg_id,
        cfg=cfg,
        overrideConfig=override_config,
        formatLogs=format_logs,
        withResult=with_result,
        runSettings=create_run_settings(
            callback_url=callback_url,
            debug_mode=debug_mode,
            core_manage=core_manage,
            single_request=single_request,
            wait_runs=wait_runs,
            profile_mode=profile_mode,
            scale_info=scale_info,
            component=component,
            policy=policy,
            restrictions=restrictions,
        )
    )


def create_cfg_struct(
    collections: Optional[Dict[Alias.Id, Union[Alias.Id, Dict[str, Any]]]] = None,
    different: Optional[Dict[Alias.Id, Alias.Id]] = None,
    schemes_aliases: Optional[Dict[Alias.Id, Alias.Id]] = None,
    msg_url: str = DEFAULT_MSG_URL,
    init_apps_update: Optional[Dict[str, bool]] = None,
    app_settings: Optional[List[AppSettings]] = None,
    app_cfg_extension: Optional[Dict[str, Union[Dict[str, Any], Alias.Json]]] = None,   # taskId$appId -> app_cfg json
    email: Optional[str] = None,
) -> Cfg:
    collections = {} if collections is None else collections
    different = {} if different is None else different
    schemes_aliases = {} if schemes_aliases is None else schemes_aliases
    init_apps_update = {} if init_apps_update is None else init_apps_update
    app_settings = [] if app_settings is None else app_settings
    app_cfg_extension_fixed = {}
    if app_cfg_extension is not None:
        for task_id_app_id, cfg in app_cfg_extension.items():
            app_cfg_extension_fixed[task_id_app_id] = to_json(cfg)
    return Cfg(
        collections=collections,
        different=different,
        schemes_aliases=schemes_aliases,
        msg_url=msg_url,
        init_apps_update=init_apps_update,
        app_settings=app_settings,
        app_cfg_extension=app_cfg_extension_fixed,
        email=email,
    )


def base_settings(
    memory_request: Optional[int] = None,
    memory_limit: Optional[int] = None,
    cpu_request: Optional[int] = None,
    cpu_limit: Optional[int] = None,
    storage_request: Optional[int] = None,
    storage_limit: Optional[int] = None,
) -> str:
    return BasePlatformSettings(
        memoryRequest=memory_request,
        memoryLimit=memory_limit,
        cpuRequest=cpu_request,
        cpuLimit=cpu_limit,
        storageRequest=storage_request,
        storageLimit=storage_limit,
    ).model_dump_json()



import urllib
from typing import Dict, Optional

import aiohttp

from malevich_coretools.batch import DefferOperationInternal
from malevich_coretools.secondary.helpers import bool_to_str

# const
API_VERSION = "api/v1"
HEADERS = {'Content-type': 'application/json', 'Accept': 'application/json', 'User-Agent': 'malevich user agent'}
HEADERS_RAW = {'Content-type': 'text/plain; base64', 'Accept': 'application/json', 'User-Agent': 'malevich user agent'}
SLEEP_TIME = 0.1
LONG_SLEEP_TIME = 1             # second
WAIT_RESULT_TIMEOUT = 60 * 60   # hour
AIOHTTP_TIMEOUT = aiohttp.ClientTimeout(total=60 * 10) # 10 min
AIOHTTP_TIMEOUT_MINI = aiohttp.ClientTimeout(total=60 * 5) # 5 min
POSSIBLE_APPS_PLATFORMS = {"base", "vast"}
SCHEME_PATTERN = r"[a-zA-Z_]\w+"

# endpoints
def with_wait(url, wait) -> str:
    return url if wait is None else f"{url}?wait={bool_to_str(wait)}"    # always first

def with_key_values(url: str, key_values: Dict[str, Optional[str]]) -> str:
    sep = "?"
    for key, value in key_values.items():
        if value is not None:
            if isinstance(value, DefferOperationInternal):
                value = urllib.parse.quote(str(value), safe='')
            elif isinstance(value, str):
                value = urllib.parse.quote(value, safe='')
            url = f"{url}{sep}{key}={value}"
            if sep == "?":
                sep = "&"
    return url

## DocsController
DOCS_MAIN = f"{API_VERSION}/docs"
DOCS = lambda wait: with_wait(f"{DOCS_MAIN}/", wait)
DOCS_ID = lambda id, wait: with_wait(f"{DOCS_MAIN}/{id}", wait)
DOCS_NAME = lambda name, wait: with_wait(f"{DOCS_MAIN}/name/{name}", wait)

## CollectionsController
COLLECTIONS_MAIN = f"{API_VERSION}/collections"
COLLECTIONS = lambda wait: with_wait(f"{COLLECTIONS_MAIN}/", wait)
COLLECTIONS_IDS_NAME = lambda name, operation_id, run_id: with_key_values(f"{COLLECTIONS_MAIN}/ids/name/{urllib.parse.quote(str(name), safe='')}", {"operationId": operation_id, "runId": run_id})
COLLECTIONS_NAME = lambda name, operation_id, run_id, offset, limit: with_key_values(f"{COLLECTIONS_MAIN}/name/{urllib.parse.quote(str(name), safe='')}", {"operationId": operation_id, "runId": run_id, "offset": offset, "limit": limit})
COLLECTIONS_IDS_GROUP_NAME = lambda name, operation_id, run_id: with_key_values(f"{COLLECTIONS_MAIN}/ids/groupName/{urllib.parse.quote(str(name), safe='')}", {"operationId": operation_id, "runId": run_id})
COLLECTIONS_GROUP_NAME = lambda name, operation_id, run_id: with_key_values(f"{COLLECTIONS_MAIN}/groupName/{urllib.parse.quote(str(name), safe='')}", {"operationId": operation_id, "runId": run_id})
COLLECTIONS_ID = lambda id, offset, limit: with_key_values(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}", {"offset": offset, "limit": limit})
COLLECTIONS_ID_MODIFY = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)
COLLECTIONS_ID_S3 = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/s3/{urllib.parse.quote(str(id), safe='')}", wait)
COLLECTIONS_ID_ADD = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/add", wait)
COLLECTIONS_ID_COPY = lambda id, full_copy, wait: with_key_values(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/copy", {"wait": None if wait is None else bool_to_str(wait), "fullCopy": None if full_copy is None else bool_to_str(full_copy)})
COLLECTIONS_ID_DEL = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/del", wait)
COLLECTIONS_DATA = lambda wait: with_wait(f"{COLLECTIONS_MAIN}/data", wait)
COLLECTIONS_DATA_ID = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/data/{urllib.parse.quote(str(id), safe='')}", wait)
COLLECTIONS_APPLY_SCHEME = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/applyScheme", wait)
COLLECTIONS_FIX_SCHEME = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/fixScheme", wait)
COLLECTIONS_UNFIX_SCHEME = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/unfixScheme", wait)
COLLECTIONS_METADATA = lambda id, wait: with_wait(f"{COLLECTIONS_MAIN}/{urllib.parse.quote(str(id), safe='')}/metadata", wait)

## CollectionObjectsController
COLLECTION_OBJECTS_MAIN = f"{API_VERSION}/collectionObjects"
COLLECTION_OBJECTS_ALL_GET = lambda path, recursive: with_key_values(f"{COLLECTION_OBJECTS_MAIN}/all", {"path": path, "recursive": recursive})
COLLECTION_OBJECTS_ALL = lambda wait: with_wait(f"{COLLECTION_OBJECTS_MAIN}/all", wait)
COLLECTION_OBJECTS_PATH = lambda path, wait, zip: with_key_values(f"{COLLECTION_OBJECTS_MAIN}/", {"path": path, "wait": None if wait is None else bool_to_str(wait), "zip": None if zip is None else bool_to_str(zip)})
COLLECTION_OBJECTS_PRESIGN_PUT = lambda path, callback_url, expires_in, wait: with_key_values(f"{COLLECTION_OBJECTS_MAIN}/presign/put", {"path": path, "callback_url": callback_url, "expiresIn": expires_in, "wait": bool_to_str(wait)})
COLLECTION_OBJECTS_PRESIGN_GET = lambda path, callback_url, expires_in, wait: with_key_values(f"{COLLECTION_OBJECTS_MAIN}/presign/get", {"path": path, "callback_url": callback_url, "expiresIn": expires_in, "wait": bool_to_str(wait)})
COLLECTION_OBJECTS_PRESIGN = lambda signature, zip: with_key_values(f"{COLLECTION_OBJECTS_MAIN}/presign", {"signature": signature, "zip": None if zip is None else bool_to_str(zip)})

## EndpointController
ENDPOINTS_MAIN = f"{API_VERSION}/endpoints"
ENDPOINTS = lambda hash, wait: with_wait(f"{ENDPOINTS_MAIN}/{urllib.parse.quote(str(hash), safe='')}", wait)
ENDPOINTS_ALL = lambda wait: with_wait(f"{ENDPOINTS_MAIN}/all", wait)
ENDPOINTS_RUN = lambda hash: f"{ENDPOINTS_MAIN}/run/{urllib.parse.quote(str(hash), safe='')}"
ENDPOINTS_CREATE = lambda wait: with_wait(f"{ENDPOINTS_MAIN}/create", wait)
ENDPOINTS_UPDATE = lambda wait: with_wait(f"{ENDPOINTS_MAIN}/update", wait)
ENDPOINTS_PAUSE = lambda hash, wait: with_wait(f"{ENDPOINTS_MAIN}/pause/{urllib.parse.quote(str(hash), safe='')}", wait)
ENDPOINTS_RESUME = lambda hash, wait: with_wait(f"{ENDPOINTS_MAIN}/resume/{urllib.parse.quote(str(hash), safe='')}", wait)

## SchemeController
SCHEMES_MAIN = f"{API_VERSION}/schemes"
SCHEMES = lambda wait: with_wait(f"{SCHEMES_MAIN}/", wait)
SCHEMES_ID = lambda id, wait: with_wait(f"{SCHEMES_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)
SCHEMES_ID_RAW = lambda id: f"{SCHEMES_MAIN}/{urllib.parse.quote(str(id), safe='')}/raw"
SCHEMES_MAPPING = lambda wait: with_wait(f"{SCHEMES_MAIN}/mapping", wait)
SCHEMES_MAPPING_IDS = lambda from_id, to_id: f"{SCHEMES_MAIN}/mapping/{urllib.parse.quote(str(from_id), safe='')}/{urllib.parse.quote(str(to_id), safe='')}"

## CommonController
CHECK = ""
PING = "ping"
# COMMON_MAIN = f"{API_VERSION}/common"
# MAPPING = lambda wait: with_wait(f"{COMMON_MAIN}/mapping", wait)
# MAPPING_ID = lambda id, wait: with_wait(f"{COMMON_MAIN}/mapping/{urllib.parse.quote(str(id), safe='')}", wait)
# COMMON_ALL = lambda wait: with_wait(f"{COMMON_MAIN}/all", wait)

## UserShareController
SHARE_MAIN = f"{API_VERSION}/share"
SHARE = lambda wait: with_wait(f"{SHARE_MAIN}/", wait)
SHARE_COLLECTION_ID = lambda id, wait: with_wait(f"{SHARE_MAIN}/collection/{urllib.parse.quote(str(id), safe='')}", wait)
SHARE_SCHEME_ID = lambda id, wait: with_wait(f"{SHARE_MAIN}/scheme/{urllib.parse.quote(str(id), safe='')}", wait)
SHARE_USER_APP_ID = lambda id, wait: with_wait(f"{SHARE_MAIN}/userApp/{urllib.parse.quote(str(id), safe='')}", wait)
SHARE_LOGIN = lambda login: f"{SHARE_MAIN}/login/{urllib.parse.quote(str(login), safe='')}"
SHARE_ALL = lambda wait: with_wait(f"{SHARE_MAIN}/all", wait)

## RegistrationController
REGISTER_MAIN = f"{API_VERSION}/register"
REGISTER = f"{REGISTER_MAIN}/"
REGISTER_LOGIN = lambda login, wait: with_wait(f"{REGISTER_MAIN}/login/{urllib.parse.quote(str(login), safe='')}", wait)
REGISTER_ALL = f"{REGISTER_MAIN}/all"

## UserAppsController
USER_APPS_MAIN = f"{API_VERSION}/userApps"
USER_APPS = lambda wait: with_wait(f"{USER_APPS_MAIN}/", wait)
USER_APPS_REAL_IDS = f"{USER_APPS_MAIN}/realIds"
USER_APPS_MAP_IDS = f"{USER_APPS_MAIN}/mapIds"
USER_APPS_MAP_ID = lambda id: f"{USER_APPS_MAIN}/mapIds/{urllib.parse.quote(str(id), safe='')}"
USER_APPS_ID = lambda id, wait: with_wait(f"{USER_APPS_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)
USER_APPS_REAL_ID = lambda id: f"{USER_APPS_MAIN}/realIds/{urllib.parse.quote(str(id), safe='')}"

## UserTasksController
USER_TASKS_MAIN = f"{API_VERSION}/userTasks"
USER_TASKS = lambda wait: with_wait(f"{USER_TASKS_MAIN}/", wait)
USER_TASKS_REAL_IDS = f"{USER_TASKS_MAIN}/realIds"
USER_TASKS_MAP_IDS = f"{USER_TASKS_MAIN}/mapIds"
USER_TASKS_MAP_ID = lambda id: f"{USER_TASKS_MAIN}/mapIds/{urllib.parse.quote(str(id), safe='')}"
USER_TASKS_ID = lambda id, wait: with_wait(f"{USER_TASKS_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)
USER_TASKS_REAL_ID = lambda id: f"{USER_TASKS_MAIN}/realIds/{urllib.parse.quote(str(id), safe='')}"

## UserPipelinesController
USER_PIPELINES_MAIN = f"{API_VERSION}/userPipelines"
USER_PIPELINES = lambda wait: with_wait(f"{USER_PIPELINES_MAIN}/", wait)
USER_PIPELINES_REAL_IDS = f"{USER_PIPELINES_MAIN}/realIds"
USER_PIPELINES_MAP_IDS = f"{USER_PIPELINES_MAIN}/mapIds"
USER_PIPELINES_MAP_ID = lambda id: f"{USER_PIPELINES_MAIN}/mapIds/{urllib.parse.quote(str(id), safe='')}"
USER_PIPELINES_ID = lambda id, wait: with_wait(f"{USER_PIPELINES_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)
USER_PIPELINES_REAL_ID = lambda id: f"{USER_PIPELINES_MAIN}/realIds/{urllib.parse.quote(str(id), safe='')}"

## UserCfgsController
USER_CFGS_MAIN = f"{API_VERSION}/userCfgs"
USER_CFGS = lambda wait: with_wait(f"{USER_CFGS_MAIN}/", wait)
USER_CFGS_REAL_IDS = f"{USER_CFGS_MAIN}/realIds"
USER_CFGS_MAP_IDS = f"{USER_CFGS_MAIN}/mapIds"
USER_CFGS_MAP_ID = lambda id: f"{USER_CFGS_MAIN}/mapIds/{urllib.parse.quote(str(id), safe='')}"
USER_CFGS_ID = lambda id, wait: with_wait(f"{USER_CFGS_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)
USER_CFGS_REAL_ID = lambda id: f"{USER_CFGS_MAIN}/realIds/{urllib.parse.quote(str(id), safe='')}"

## OperationResultsController
OPERATION_RESULTS_MAIN = f"{API_VERSION}/operationResults"
OPERATION_RESULTS = lambda wait: with_wait(f"{OPERATION_RESULTS_MAIN}/", wait)
OPERATION_RESULTS_ID = lambda id, wait: with_wait(f"{OPERATION_RESULTS_MAIN}/{urllib.parse.quote(str(id), safe='')}", wait)

## TempRunController
TEMP_RUN_MAIN = f"{API_VERSION}/run"
TEMP_RUN_CONDITION = lambda operationId: f"{TEMP_RUN_MAIN}/condition/{urllib.parse.quote(str(operationId), safe='')}"
TEMP_RUN_ACTIVE_RUNS = f"{TEMP_RUN_MAIN}/activeRuns"
TEMP_RUN_MAIN_TASK_CFG = lambda operationId: f"{TEMP_RUN_MAIN}/mainTaskCfg/{urllib.parse.quote(str(operationId), safe='')}"
TEMP_RUN_MAIN_PIPELINE_CFG = lambda operationId: f"{TEMP_RUN_MAIN}/mainPipelineCfg/{urllib.parse.quote(str(operationId), safe='')}"
TEMP_RUN_OPERATIONS_IDS = lambda taskId, cfgId: f"{TEMP_RUN_MAIN}/operationsIds/{urllib.parse.quote(str(taskId), safe='')}" if cfgId is None else f"{TEMP_RUN_MAIN}/operationsIds/{urllib.parse.quote(str(taskId), safe='')}/{urllib.parse.quote(str(cfgId), safe='')}"

## AdminController
ADMIN_MAIN = f"{API_VERSION}/admin"
ADMIN_RUNS = f"{ADMIN_MAIN}/runs"
ADMIN_RUNS_INFO = f"{ADMIN_MAIN}/runs/info"
ADMIN_SUPERUSER = f"{ADMIN_MAIN}/superuser"

## ManagerController
MANAGER_MAIN = f"{API_VERSION}/manager"
MANAGER_LOGS = f"{MANAGER_MAIN}/logs"
MANAGER_CLICKHOUSE_ALL = f"{MANAGER_MAIN}/clickhouse"
MANAGER_CLICKHOUSE_ID = lambda operationId: f"{MANAGER_MAIN}/clickhouse/{urllib.parse.quote(str(operationId), safe='')}"
MANAGER_DAG_KEY_VALUE = lambda wait: with_wait(f"{MANAGER_MAIN}/dagKeyValue", wait)
MANAGER_DAG_KEY_VALUE_OPERATION_ID = lambda operationId: f"{MANAGER_MAIN}/dagKeyValue/{urllib.parse.quote(str(operationId), safe='')}"
MANAGER_APP_INFO = lambda appId: f"{MANAGER_MAIN}/appInfo/{urllib.parse.quote(str(appId), safe='')}"
MANAGER_APP_INFO_REAL_ID = lambda appId: f"{MANAGER_MAIN}/appInfo/realId/{urllib.parse.quote(str(appId), safe='')}"
MANAGER_IMAGE_INFO = f"{MANAGER_MAIN}/imageInfo"
MANAGER_TASK_SCHEDULES = f"{MANAGER_MAIN}/task/schedules"
MANAGER_TASK = lambda wait: with_wait(f"{MANAGER_MAIN}/task", wait)
MANAGER_TASK_RUN = lambda wait: with_wait(f"{MANAGER_MAIN}/task/run", wait)
MANAGER_TASK_UNSCHEDULE = lambda wait: with_wait(f"{MANAGER_MAIN}/task/unschedule", wait)
MANAGER_TASK_STOP = lambda wait: with_wait(f"{MANAGER_MAIN}/task/stop", wait)
MANAGER_TASK_STOP_ALL = lambda wait: with_wait(f"{MANAGER_MAIN}/task/stopAll", wait)
MANAGER_TASK_RESUME = lambda wait: with_wait(f"{MANAGER_MAIN}/task/resume", wait)
MANAGER_TASK_PAUSE = lambda wait: with_wait(f"{MANAGER_MAIN}/task/pause", wait)
MANAGER_PIPELINE = lambda wait: with_wait(f"{MANAGER_MAIN}/pipeline", wait)
MANAGER_APP_STOP = lambda wait: with_wait(f"{MANAGER_MAIN}/app/stop", wait)
MANAGER_APP_RESUME = lambda wait: with_wait(f"{MANAGER_MAIN}/app/resume", wait)
MANAGER_APP_PAUSE = lambda wait: with_wait(f"{MANAGER_MAIN}/app/pause", wait)

## LimitsController
LIMITS_MAIN = f"{API_VERSION}/limits"
LIMITS = lambda wait: with_wait(f"{LIMITS_MAIN}/", wait)
LIMITS_USER = lambda wait: with_wait(f"{LIMITS_MAIN}/user", wait)

## HandlerUrlController
HANDLER_URL_MAIN = f"{API_VERSION}/handlerUrls"
HANDLER_URL = lambda url, wait: with_key_values(f"{HANDLER_URL_MAIN}/", {"url": url, "wait": bool_to_str(wait)})

## AnalyticsController
ANALYTICS_MAIN = f"{API_VERSION}/analytics"
ANALYTICS = lambda wait: with_wait(f"{ANALYTICS_MAIN}/", wait)
ANALYTICS_MANY = lambda wait: with_wait(f"{ANALYTICS_MAIN}/many", wait)
ANALYTICS_ID = lambda id, wait: with_wait(f"{ANALYTICS_MAIN}/{id}", wait)
ANALYTICS_NAME = lambda name, wait: with_wait(f"{ANALYTICS_MAIN}/name/{name}", wait)

### Kafka
KAFKA_SEND = f"{MANAGER_MAIN}/kafkaMsg"

## BatchController
BATCH_MAIN = f"{API_VERSION}/batch"
BATCH = f"{BATCH_MAIN}/"



from enum import IntEnum
from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from malevich_coretools.abstract.abstract import (  # noqa: F401
    AdminRunInfo,
    Alias,
    JsonImage,
    Restrictions,
    ScaleInfo,
    Schedule,
    TaskComponent,
    TaskPolicy,
)
from malevich_coretools.abstract.operations import (
    And,
    BoolOp,
    InternalAnd,
    InternalBoolOp,
    InternalNot,
    InternalOr,
    Not,
    Or,
    Value,
)


class PullCollectionPolicy(IntEnum):
    INIT = 0
    IF_NOT_EXIST = 1
    FORCE_RELOAD = 2
    FORCE_RELOAD_ALL = 3


class BaseArgument(BaseModel):
    id: Optional[str] = None                            # from - bindProcessorId
    indices: Optional[List[int]] = None                 # full if None
    # or
    collectionName: Optional[str] = None                # get id (or obj path) from cfg by it
    # or
    collectionId: Optional[str] = None                  # hardcode collection with id (or obj path)

    def validation(self) -> None:
        assert (self.id is not None) + (self.collectionName is not None) + (self.collectionId is not None) == 1, "one way of constructing the argument must be chosen"


class Argument(BaseArgument):
    group: Optional[List[BaseArgument]] = None          # for constructed dfs, sink
    conditions: Optional[Dict[str, bool]] = None        # valid only for alternative, bindConditionId -> value, must be specified explicitly, then it will be derived from the pipeline structure

    def validation(self) -> None:
        if self.group is not None:
            assert self.id is None and self.collectionName is None and self.collectionId is None, "one way of constructing the argument must be chosen"
            for subarg in self.group:
                subarg.validation()
        else:
            assert (self.id is not None) + (self.collectionName is not None) + (self.collectionId is not None) == 1, "one way of constructing the argument must be chosen"


class AlternativeArgument(BaseArgument):
    group: Optional[List[BaseArgument]] = None          # for constructed dfs, sink
    alternative: Optional[List[Argument]] = None        # if set - should be only one valid argument with conditions

    def validation(self) -> None:
        if self.group is not None:
            assert self.id is None and self.collectionName is None and self.collectionId is None, "one way of constructing the argument must be chosen"
            for subarg in self.group:
                subarg.validation()
        elif self.alternative is not None:
            for alt_arg in self.alternative:
                if alt_arg.group is not None:
                    assert alt_arg.id is None and alt_arg.collectionName is None and alt_arg.collectionId is None, "one way of constructing the argument must be chosen"
                    for subarg in alt_arg.group:
                        subarg.validation()
                else:
                    assert (self.id is not None) + (self.collectionName is not None) + (self.collectionId is not None) == 1, "one way of constructing the argument must be chosen"
        else:
            assert (self.id is not None) + (self.collectionName is not None) + (self.collectionId is not None) == 1, "one way of constructing the argument must be chosen"


class AppEntity(BaseModel):
    cfg: Optional[str] = None                           # local cfg for processor/condition
    arguments: Dict[str, AlternativeArgument] = {}      # TODO or List[Argument]?
    conditions: Optional[Union[Dict[str, bool], List[Dict[str, bool]]]] = None          # condition bindId to it result (list - any variant of them)
    conditionsStructure: Optional[Union[Not, And, Or, InternalNot, InternalAnd, InternalOr, Value]] = None  # set BoolOp, it transform to internal before send

    loopArguments: Optional[Dict[str, AlternativeArgument]] = None  # other calls, TODO or List[Argument]?, problems
    loopConditions: Optional[Union[Dict[str, bool], List[Dict[str, bool]]]] = None          # condition bindId to it result for loop (list - any variant of them)
    loopConditionsStructure: Optional[Union[Not, And, Or, InternalNot, InternalAnd, InternalOr, Value]] = None  # set BoolOp, it transform to internal before send

    image: JsonImage
    platform: str = "base"
    platformSettings: Optional[str] = None

    def internal(self) -> None:
        assert self.conditions is None or self.conditionsStructure is None, "should be set not more, than one of (conditions, conditionsStructure)"
        assert self.loopConditions is None or self.loopConditionsStructure is None, "should be set not more, than one of (loopConditions, loopConditionsStructure)"

        if isinstance(self.conditions, Dict):
            self.conditions = [self.conditions]
        if isinstance(self.conditionsStructure, BoolOp):
            self.conditionsStructure = self.conditionsStructure.internal()
        if isinstance(self.loopConditions, Dict):
            self.loopConditions = [self.loopConditions]
        if isinstance(self.loopConditionsStructure, BoolOp):
            self.loopConditionsStructure = self.loopConditionsStructure.internal()

    def simplify(self) -> None:
        if isinstance(self.conditionsStructure, InternalBoolOp):
            self.conditionsStructure = self.conditionsStructure.simplify()
        if isinstance(self.loopConditionsStructure, InternalBoolOp):
            self.loopConditionsStructure = self.loopConditionsStructure.simplify()


class Processor(AppEntity):
    processorId: str
    outputId: Optional[str] = None


class Condition(AppEntity):                             # at least one of (true, false) set
    conditionId: str


class Result(BaseModel):                                # save collection (only processor)
    name: str
    index: Optional[int] = None


class Pipeline(BaseModel):
    pipelineId: str
    processors: Dict[str, Processor] = {}               # bindProcessorId to Processor
    conditions: Dict[str, Condition] = {}               # bindConditionId to Condition
    results: Dict[str, List[Result]] = {}               # bindProcessorId to results
    pullCollectionPolicy: PullCollectionPolicy = PullCollectionPolicy.IF_NOT_EXIST

    def internal(self) -> 'Pipeline':
        for proc in self.processors.values():
            proc.internal()
        for cond in self.conditions.values():
            cond.internal()
        return self

    def simplify(self) -> 'Pipeline':
        for proc in self.processors.values():
            proc.simplify()
        for cond in self.conditions.values():
            cond.simplify()
        return self


class MainPipelineCfg(BaseModel):
    operationId: Alias.Id
    pipelineId: Alias.Id
    processors: Dict[str, Processor]
    conditions: Dict[str, Condition]
    results: Dict[str, List[Result]]
    pullCollectionPolicy: PullCollectionPolicy
    cfg: str
    infoUrl: Optional[str] = None
    debugMode: bool
    coreManage: bool
    kafkaMode: bool
    singleRequest: bool
    tlWithoutData: Optional[int] = None
    waitRuns: bool
    profileMode: Optional[str] = None
    withLogs: bool
    component: TaskComponent
    policy: TaskPolicy
    schedule: Optional[Schedule] = None
    restrictions: Optional[Restrictions] = None
    scaleInfo: List[ScaleInfo]
    schemesNames: List[str]
    login: Optional[Alias.Login] = None
    withListener: bool
    kafkaModeUrl: Optional[str] = None
    run: bool
    synthetic: bool = False


class PipelineInfo(BaseModel):
    pipelineId: Alias.Id
    processors: Dict[str, Processor]
    conditions: Dict[str, Condition]
    results: Dict[str, List[Result]]
    cfg: Alias.Json
    login: Alias.Login


class AdminRunPipelineInfo(BaseModel):
    operationId: Alias.Id
    pipelineInfo: PipelineInfo
    cfgId: Alias.Id


class AdminRunsInfo(BaseModel):
    tasks: List[AdminRunInfo]
    pipelines: List[AdminRunPipelineInfo]
